#include <algo/blast/core/gpu_cpu_common.h>
#include "gpu_blastp_kernel.h"
#define CHARSIZE 5

//shared and constant memory
extern __shared__ int array[];
__constant__ char d_Query_const[40960];
__constant__ int d_x_dropoff_const[500];
__constant__ int d_cutoff_score_const[500];
__constant__ int d_query_offset_const[500];
__constant__ int Group_size_const;


/**
 * Each thread scans on of the subjet sequences and
 * it returns the score and coordinats of each successful alignment
 * @param d_Database2Dpadded                   the formatted sequence database[in]
 * @param d_RepeatedSubstitutionMatrix         multiple copies of the substitution matrix (one for each block for parallel loading from global to shared) [in]
 * @param SubstitutionMatrix_length            the substitution matrix [in]
 * @param d_RepeatedPV                         multiple copies of the presence vector (one for each block for parallel loading from global to shared) [in]
 * @param d_ThickBackbone                      ThickBackbone as taken from NCBI_BLAST [in]
 * @param d_overflow                           overflow array as taken from NCIB_BLAST [in]
 * @param *d_RepeatedSequenceMaxlength_vector  multiple copies of the vector that has the maximum length for each group [in]
 * @param SequenceMaxlength_vector_stride      the stride of the array that holds the maximum lengths of each group [in]
 * @param NumSequences                         number of sequences the GPU works on [in]
 * @param pv_length                            length of the presence vector [in]
 * @param Query_length                         length of the query [in]
 * @param diag_mask                            mask for the diagonal array [in]
 * @param diag_array_length                    length of the diagonal array used by each thread [in]
 * @param num_quries                           number of queries [in]
 * @param word_length                          word length [in]
 * @param charsize                             char size of the lookup table [in]
 * @param index_mask                           mask for the iniatialization of the index [in]
 * @param window                               window for the two-hit alignment [in]
 * @param diag_offset                          diagonal offset [in]
 * @param d_diag_array                         allocated global memory used for storing the diagonal arrays [in]
 * @param d_Hits returns                       returns hits found for each sequence [out]
 * @param d_GPUBlastInitHitList                returns the statistics of the successful extensions [out]
*/

__global__ void
GPU_BLASTP_kernelTwoHit( int* d_Database2Dpadded,
                         const int* d_RepeatedSubstitutionMatrix,
                         const int SubstitutionMatrix_length,
                         const PV_ARRAY_TYPE* d_RepeatedPV,
                         const AaLookupSmallboneCell* d_ThickBackbone,
                         const unsigned short* d_overflow,
                         const int* d_RepeatedSequence_length_vector,
                         const int Sequence_length_vector_stride,
                         const int NumSequences,
                         const int pv_length,
                         const int Query_length,
                         const int diag_mask,
                         const int diag_array_length,
                         const int num_queries,
                         const int word_length,
                         const int index_mask,
                         const int window,
                         int diag_offset,
                         short* d_diag_array,
                         int* d_Hits,
                         GPUBlastInitHitList* d_GPUBlastInitHitList,
                         int hit_list_base,
                         unsigned int total_hit_list_length)
{
    int bx = blockIdx.x;
    int tx = threadIdx.x;
    int blockDimx = blockDim.x;

    //total threads in the card
    int total_threads = blockDim.x*gridDim.x;

    //total room for results per thread
    int left_room = total_hit_list_length/total_threads;

    int current_subject_offset = left_room*( bx*blockDimx + (tx&0xfe0))+ left_room * 2* (tx&0xf);

    left_room *= 2;

    int kept_results_count_for_current_subject;

    if( 0==(tx&0x10) )
    {
        d_GPUBlastInitHitList = &(d_GPUBlastInitHitList[current_subject_offset]);
        hit_list_base += current_subject_offset;
    }
    else
    {
        d_GPUBlastInitHitList = &(d_GPUBlastInitHitList[current_subject_offset+left_room-1]);
        hit_list_base += current_subject_offset+left_room-1;
    }



    int* SubstitutionMatrix_shared_int = (int*)array;
    int* Sequence_length_vector_shared = (int*)&SubstitutionMatrix_shared_int[SubstitutionMatrix_length>>2];//The substitution matrix is 640 bytes or 160 integers long
    int* result_space_used = (int*)  &Sequence_length_vector_shared[Sequence_length_vector_stride];
    char* Sequence_AA = (char*) &result_space_used[blockDimx];

    PV_ARRAY_TYPE* pv_shared = (PV_ARRAY_TYPE*) &Sequence_AA[3* blockDimx * 4];

    //pv array from global to shared memory
    int pv_Index_global = bx*pv_length + tx;
    int pv_Index_shared = tx;
    while( pv_Index_shared < pv_length )
    {
        pv_shared[ pv_Index_shared ] = d_RepeatedPV[ pv_Index_global ];

        pv_Index_global += blockDimx;
        pv_Index_shared += blockDimx;
    }

    __syncthreads();

    //Substitution Matrix from global to shared memory
    int SubstitutionMatrix_Index_global = bx*(SubstitutionMatrix_length>>2) + tx;
    int SubstitutionMatrix_Index_shared = tx;
    while( SubstitutionMatrix_Index_shared < (SubstitutionMatrix_length>>2) )//SubstitutionMatrix_length is a multiple of 64 which includes some padded zeros so there is some waste of memory space. Can it be recovered?
    {
        SubstitutionMatrix_shared_int[ SubstitutionMatrix_Index_shared ] = d_RepeatedSubstitutionMatrix[ SubstitutionMatrix_Index_global ];

        SubstitutionMatrix_Index_global += blockDimx;
        SubstitutionMatrix_Index_shared += blockDimx;
    }

    __syncthreads();

    //SequenceMaxlength_vector from global to shared memory
    int Sequence_length_vector_Index_global = bx*Sequence_length_vector_stride + tx;
    int Sequence_length_vector_Index_shared = tx;
    while( Sequence_length_vector_Index_shared < Sequence_length_vector_stride )
    {
        Sequence_length_vector_shared[ Sequence_length_vector_Index_shared ] = d_RepeatedSequence_length_vector[ Sequence_length_vector_Index_global ];

        Sequence_length_vector_Index_global += blockDimx;
        Sequence_length_vector_Index_shared += blockDimx;
    }

    __syncthreads();


    char* SubstitutionMatrix_shared_char = (char*)SubstitutionMatrix_shared_int;

    int Sequence = (bx * blockDimx + tx);

    GPUBlastInitHitList data;

    int Group_size = gridDim.x*blockDimx;
    int diag_array_base = ((bx*blockDimx+tx)*diag_array_length);

    Sequence_AA = &Sequence_AA[tx*12];
    d_diag_array = &d_diag_array[diag_array_base];
    d_Database2Dpadded = &(d_Database2Dpadded[bx*blockDimx + tx+ blockDim.x*gridDim.x ]);


    int S_AA_buffer;

    while( Sequence < NumSequences )
    {
        kept_results_count_for_current_subject=0;
        int Hits = 0;
        int s_last_off = 0;
        int hits_extended = 0;
        int successful_extensions = 0;
        int Sequence_length = Sequence_length_vector_shared[ Sequence/Group_size ];

        ((int*) Sequence_AA)[0] = d_Database2Dpadded[ 0 ];
        S_AA_buffer = d_Database2Dpadded[ Group_size ];

        int index = (Sequence_AA[ 0 ] << CHARSIZE) |  Sequence_AA[ 1 ];

        for(int s_off = 0; (s_off <= Sequence_length - word_length); ++s_off)
        {

            if ( ((s_off&0x3)==0) && (s_off < Sequence_length-4) )
            {
                ((int*)Sequence_AA)[((s_off>>2)+1)&0x1] = S_AA_buffer;
                S_AA_buffer = d_Database2Dpadded[ (((s_off>>2) + 2)*Group_size) ];
            }

            index = ((index<<CHARSIZE) | Sequence_AA[ (s_off+2) &0x7 ]) & index_mask;

            int numhits = 0;
            if( PV_TEST(pv_shared, index, PV_ARRAY_BTS) ) //check via the precence vector (PV) if this database subject word exists in the query
            {
                numhits = d_ThickBackbone[index].num_used;
                Hits += numhits;

                const unsigned short *src;
                if( numhits <= AA_HITS_PER_CELL ) //check there are less than AA_HITS_PER_CELL occurrences of this word in the query
                    src = d_ThickBackbone[index].payload.entries;
                else
                    src = &(d_overflow[d_ThickBackbone[index].payload.overflow_cursor]); //if not fetch the occurrences from the overflow array

                int query_offset, diag_coord, query_offset_buffer1, query_offset_buffer2;
                short diagonal, diagonal_buffer;
                short last_hit;
                short flag;


                query_offset_buffer1 = src[0];
                query_offset_buffer2 = src[1];
                numhits --;
                diag_coord = (query_offset_buffer1 - s_off) & diag_mask;
                diagonal_buffer = d_diag_array[diag_coord];


                for(int i = 0; i < numhits; ++i) //for each occurrence expand and calculate score
                {
                    query_offset = query_offset_buffer1;
                    diagonal = diagonal_buffer;

                    diagonal_buffer = d_diag_array[(query_offset_buffer2 - s_off) & diag_mask];
                    query_offset_buffer1= query_offset_buffer2;
                    query_offset_buffer2 = src[i+2];

                    diag_coord = (query_offset - s_off) & diag_mask;
                    last_hit = diagonal & 0x7FFF;
                    flag = diagonal & 0x8000;


                    /* If the reset bit is set, an extension just happened. */
                    if (flag)
                    {
                        /* If we've already extended past this hit, skip it. */
                        if ( (s_off + diag_offset) < last_hit)
                            continue;
                        else /* Otherwise, start a new hit. */
                            d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF;//set the MSB to zero and the rest to (subject_offset + diag_offset)

                    }
                    else     /* If the reset bit is cleared, try to start an extension. */
                    {

                        /* find the distance to the last hit on this diagonal */
                        last_hit = last_hit - diag_offset;
                        int diff = s_off - last_hit;

                        if (diff >= window)
                        {
                            /* We are beyond the window for this diagonal; start a new hit */
                            /* If the difference is less than the wordsize (i.e. last hit and this hit overlap), give up */
                            d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF; //set only the 31 LSB's (the flag is zero since we are here)
                            continue;
                        }

                        /* If the difference is less than the wordsize (i.e. last hit and this hit overlap), give up */
                        if (diff < word_length)
                            continue;

                        /* Extend this pair of hits. The extension to the left must reach the end of the first word in order for extension to the right to proceed.
                           To use the cutoff and X-drop values appropriate for this extension, the query context must first be found */
                        int k = 0;
                        for(k = num_queries-1; k >= 0; --k)
                            if( query_offset >= d_query_offset_const[k] )
                                break;


                        int right_extend;
                        s_BlastAaExtendTwoHit(&data,
                                              SubstitutionMatrix_shared_char,
                                              (d_Database2Dpadded),
                                              Sequence_length,
                                              d_Query_const,
                                              Query_length,
                                              last_hit + word_length,
                                              s_off,
                                              query_offset,
                                              d_x_dropoff_const[k],
                                              word_length,
                                              &s_last_off,&right_extend, Sequence_AA
                                             );

                        ++hits_extended;

                        /* if the hsp meets the score threshold, report it */
                        if ( data.score >= d_cutoff_score_const[k] )
                        {
                            data.offset_from_first_position = s_off - data.hsp_s;

                            /// If there is enough space to store this hit
                            if(left_room > 0)
                            {
                                if( 0==(threadIdx.x&0x10) )
                                    d_GPUBlastInitHitList[kept_results_count_for_current_subject] = data;
                                else
                                    d_GPUBlastInitHitList[0-kept_results_count_for_current_subject] = data;

                                kept_results_count_for_current_subject++;
                                left_room--;
                            }

                            successful_extensions++;


                        }

                        /* If an extension to the right happened, reset the last hit so that future hits to this diagonal must start over. */

                        if (right_extend)
                        {

                            d_diag_array[diag_coord] = ( (s_last_off - (word_length - 1) + diag_offset) & 0x7FFF ) | 0x8000;//set the 31 LSB's (the flag is zero size we are here)
                            // and set the MSB which holds the flag (this can be merged in the previous assignment)

                        }
                        else    /* Otherwise, make the present hit into the previous hit for this diagonal */

                            d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF;//set only the 31 LSB's (the flag is zero sinze we are here)

                    }/*
                        if (flag) */
                } /* end for - done with this batch of hits*/

                diagonal = diagonal_buffer;
                query_offset = query_offset_buffer1;
                diag_coord = (query_offset - s_off) & diag_mask;
                last_hit = diagonal & 0x7FFF;
                flag = diagonal & 0x8000;

                /* If the reset bit is set, an extension just happened. */
                if (flag)
                {
                    /* If we've already extended past this hit, skip it. */
                    if ( (s_off + diag_offset) < last_hit)
                        continue;
                    else /* Otherwise, start a new hit. */
                        d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF;//set the MSB to zero and the rest to (subject_offset + diag_offset)

                }
                else     /* If the reset bit is cleared, try to start an extension. */
                {

                    /* find the distance to the last hit on this diagonal */
                    last_hit = last_hit - diag_offset;
                    int diff = s_off - last_hit;

                    if (diff >= window)
                    {
                        /* We are beyond the window for this diagonal; start a new hit */
                        /* If the difference is less than the wordsize (i.e. last hit and this hit overlap), give up */
                        d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF; //set only the 31 LSB's (the flag is zero since we are here)
                        continue;
                    }

                    /* If the difference is less than the wordsize (i.e. last hit and this hit overlap), give up */
                    if (diff < word_length)
                        continue;

                    /* Extend this pair of hits. The extension to the left must reach the end of the first word in order for extension to the right to proceed.
                       To use the cutoff and X-drop values appropriate for this extension, the query context must first be found */
                    int k = 0;
                    for(k = num_queries-1; k >= 0; --k)
                        if( query_offset >= d_query_offset_const[k] )
                            break;


                    int right_extend;
                    s_BlastAaExtendTwoHit(&data,
                                          SubstitutionMatrix_shared_char,
                                          (d_Database2Dpadded),
                                          Sequence_length,
                                          d_Query_const,
                                          Query_length,
                                          last_hit + word_length,
                                          s_off,
                                          query_offset,
                                          d_x_dropoff_const[k],
                                          word_length,
                                          &s_last_off,&right_extend, Sequence_AA
                                         );

                    ++hits_extended;

                    /* if the hsp meets the score threshold, report it */
                    if ( data.score >= d_cutoff_score_const[k] )
                    {
                        data.offset_from_first_position = s_off - data.hsp_s;

                        /// If there is enough space to store this hit
                        if(left_room > 0)
                        {
                            if( 0==(threadIdx.x&0x10) )
                                d_GPUBlastInitHitList[kept_results_count_for_current_subject] = data;
                            else
                                d_GPUBlastInitHitList[0-kept_results_count_for_current_subject] = data;

                            kept_results_count_for_current_subject++;
                            left_room--;
                        }

                        successful_extensions++;


                    }

                    /* If an extension to the right happened, reset the last hit so that future hits to this diagonal must start over. */

                    if (right_extend)
                    {

                        d_diag_array[diag_coord] = ( (s_last_off - (word_length - 1) + diag_offset) & 0x7FFF ) | 0x8000;//set the 31 LSB's (the flag is zero size we are here)
                        // and set the MSB which holds the flag (this can be merged in the previous assignment)

                    }
                    else    /* Otherwise, make the present hit into the previous hit for this diagonal */

                        d_diag_array[diag_coord] = (s_off + diag_offset) & 0x7FFF;//set only the 31 LSB's (the flag is zero sinze we are here)

                } /*
                         if (flag) */

            } //if (PV_TEST...)

        }//for Sequence_length loop

//      end_of_batch:


        diag_offset += Sequence_length + window;


        int iTmp = Sequence*4;

        result_space_used[threadIdx.x] = kept_results_count_for_current_subject;
        __syncthreads();
        int iTmp2;
        if(threadIdx.x & 0x10)
            iTmp2 = result_space_used[threadIdx.x - 16];
        else
            iTmp2 = result_space_used[threadIdx.x + 16];

        //the subject has to be recalculated
        if ( (kept_results_count_for_current_subject < successful_extensions) || (left_room <  iTmp2) )
        {
            d_Hits[ iTmp+3 ] = -1;
            left_room += kept_results_count_for_current_subject;
        }
        else //the hits of the current subjct have been stored
        {
            d_Hits[ iTmp+3 ] = hit_list_base;

            if(threadIdx.x & 0x10)
            {
                d_GPUBlastInitHitList = & (d_GPUBlastInitHitList[0-successful_extensions]);
                hit_list_base -= successful_extensions;
            }
            else
            {
                d_GPUBlastInitHitList = & (d_GPUBlastInitHitList[successful_extensions]);
                hit_list_base += successful_extensions;
            }

            left_room -= iTmp2;
        }

        d_Hits[ iTmp ] = Hits;
        d_Hits[ iTmp + 1 ] = successful_extensions;
        d_Hits[ iTmp + 2 ] = hits_extended;

        iTmp = Group_size * (Sequence_length>>2);

        Sequence += Group_size;

        d_Database2Dpadded = &(d_Database2Dpadded[iTmp]);

    } //while loop
}



/** Perform a two-hit extension. Given two hits L and R, begin
 * at R and extend to the left. If we do not reach L, abort the extension.
 * Otherwise, begin at R and extend to the right.
 *
 * @param matrix          substitution matrix [in]
 * @param subject         subject sequence [in]
 * @param subject_length  subject length [in]
 * @param query           query sequence [in]
 * @param query           query length [in]
 * @param s_left_off      left subject offset [in]
 * @param s_right_off     right subject offset [in]
 * @param q_right_off     right query offset [in]
 * @param dropoff         X dropoff parameter [in]
 * @param word_size       number of letters in one word [in]
 * @return the scores and coordinates of the alignment of the hsp.
 */


static void
__device__ s_BlastAaExtendTwoHit(GPUBlastInitHitList* data,
                                 const char* matrix,
                                 const int* subject,
                                 const int subject_length,
                                 const char* query,
                                 const int query_length,
                                 const int s_left_off,
                                 int s_right_off,
                                 int q_right_off,
                                 const int dropoff,
                                 const int word_size,
                                 int* s_last_off,
                                 int* right_extend,
                                 char* AA_shared
                                )
{
    int left_d = 0, right_d = 0;       /* left and right displacements */
    int left_score = 0, right_score = 0;       /* left and right scores */
    int i, score = 0;

    /* find one beyond the position (up to word_size-1 letters to the right)
       that gives the largest starting score */
    /* Use "one beyond" to make the numbering consistent with how it's done
       for BlastAaExtendOneHit and the "Extend" functions called here. */
    for (i = 0; i < word_size; ++i)
    {

        score += matrix[ query[q_right_off + i] + (AA_shared[(s_right_off+i) &0x7]<<5) ];

        if (score > left_score)
        {
            left_score = score;
            right_d = i + 1;    /* Position is one beyond the end of the word. */
        }
    }

    q_right_off += right_d;
    s_right_off += right_d;

    right_d = 0;
    (*right_extend) = FALSE;

    (*s_last_off) = s_right_off;
    /* first, try to extend left, from the second hit to the first hit. */
    ExtendLeftReturn ReturnLeftValue = s_BlastAaExtendLeft(matrix, subject, query, s_right_off-1, q_right_off-1, dropoff, 0);
    left_score = ReturnLeftValue.maxscore;
    left_d = ReturnLeftValue.length;

    /* Extend to the right only if left extension reached the first hit. */
    ExtendRightReturn ReturnRightValue;
    if (left_d >= (s_right_off - s_left_off))
    {
        (*right_extend) = TRUE;

        ReturnRightValue = s_BlastAaExtendRight(matrix, subject, query, s_right_off, q_right_off, dropoff, left_score, subject_length, query_length);
        right_score = ReturnRightValue.maxscore;
        right_d = ReturnRightValue.length;

    }


    (*s_last_off) = ReturnRightValue.s_last_off;
    (*data).hsp_q = q_right_off - left_d;
    (*data).hsp_s = s_right_off - left_d;;
    (*data).hsp_len = left_d + right_d;
    (*data).score = MAX(left_score, right_score);

}


/*
 * Beginning at s_off and q_off in the subject and query, respectively,
 * extend to the left until the cumulative score drops by at least
 * 'dropoff', or the end of at least one sequence is reached.
 *
 * @param matrix     the substitution matrix [in]
 * @param subject    subject sequence [in]
 * @param query      query sequence [in]
 * @param s_off      subject offset [in]
 * @param q_off      query offset [in]
 * @param dropoff    the X dropoff parameter [in]
 * @param length     the length of the computed extension [out]
 * @param maxscore   the best running score from a previous extension;
 *                   the running score of the current extension must exceed
 *                   this value if the extension is to be of nonzero size [in]
 * @return The score and the length of the extension
 */
__device__ ExtendLeftReturn s_BlastAaExtendLeft(const char* matrix,
        const int* subject,
        const char* query,
        const int s_off,
        const int q_off,
        const int dropoff,
        int maxscore)
{
    int n, best_i;
    int score = maxscore;

    int subject_buffer1, subject_buffer2;
    int temp = gridDim.x*blockDim.x;
    int index = ((s_off)>>2)*temp;
    subject_buffer1 = subject[index];
    index -= temp;
    subject_buffer2 = subject[index ];
    char* subject_buffer1_ptr = (char*) &subject_buffer1;

    n = MIN(s_off, q_off);
    best_i = n + 1;

    int temp2 = ((s_off)&0x3);
    int n1=n-temp2;

    int n2 = n1>>2;
    int k = n;

    if(n>0)
    {
        char query_temp = 0, sequence_temp = 0, flag = 0;

        for(int i=temp2; i>=0; i--)
        {
            score += matrix[ (query[q_off - n+k]<<5) + subject_buffer1_ptr[i] ];

            if (score > maxscore)
            {
                maxscore = score;
                best_i = k;
            }

            if ((maxscore - score) >= dropoff)
            {
                flag = 1;
                break;
            }
            k--;
        }

        if ( (flag == 0) && (n1>0) )
        {
            for (int i = 0; i < n2; i++)
            {
                subject_buffer1=subject_buffer2;
                index=index- temp;
                subject_buffer2 = subject[index];

                //0
                sequence_temp = subject_buffer1_ptr[3];

                query_temp = query[q_off - n+k];

                score += matrix[ (query_temp<<5) + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if ((maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }

                //1
                k--;
                sequence_temp = subject_buffer1_ptr[2];

                query_temp = query[q_off - n+k];

                score += matrix[ (query_temp<<5) + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if ((maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }
                //2
                k--;

                sequence_temp = subject_buffer1_ptr[1];

                query_temp = query[q_off - n+k];

                score += matrix[ (query_temp<<5) + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if ((maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }

                //3
                k--;
                sequence_temp = subject_buffer1_ptr[0];

                query_temp = query[q_off - n+k];

                score += matrix[ (query_temp<<5) + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if ((maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }
                k--;
            }

            if(0==flag)
            {
                subject_buffer1 = subject_buffer2;
                for(int j=0; j<(n1 &0x3); j++)
                {
                    sequence_temp = subject_buffer1_ptr[3-j];

                    query_temp = query[q_off - n+k];

                    score += matrix[ (query_temp<<5) + sequence_temp ];

                    if (score > maxscore)
                    {
                        maxscore = score;
                        best_i = k;
                    }

                    if ((maxscore - score) >= dropoff)
                        break;

                    k--;

                }
            }
        }
    }

    ExtendLeftReturn ReturnValues;
    ReturnValues.maxscore = maxscore;
    ReturnValues.length = n - best_i + 1;
    return ReturnValues;

}




/*
 * Beginning at s_off and q_off in the subject and query, respectively,
 * extend to the right until the cumulative score drops by at least
 * 'dropoff', or the end of at least one sequence is reached.
 *
 * @param matrix                      the substitution matrix [in]
 * @param subject                     subject sequence [in]
 * @param query                       query sequence [in]
 * @param s_off                       subject offset [in]
 * @param q_off                       query offset [in]
 * @param dropoff                     the X dropoff parameter [in]
 * @param maxscore                    the best running score from a previous extension;
 *                                    the running score of the current extension must exceed
 *                                    this value if the extension is to be of nonzero size [in]
 * @param Sequence_length             the length of the subject sequence [in]
 * @param Query_length                the length of the query sequence [in]
 * @return The score, lenght, and coordinates of the extension
 */
__device__ ExtendRightReturn s_BlastAaExtendRight(const char* matrix,
        const int* subject,
        const char* query,
        const int s_off,
        const int q_off,
        const int dropoff,
        int maxscore,
        const int Sequence_length,
        const int Query_length)
{
    int n, best_i = -1;
    int score = maxscore;

    int subject_buffer1, subject_buffer2;
    int temp = gridDim.x*blockDim.x;
    int index = ((s_off)>>2)*temp;
    subject_buffer1 = subject[index];
    index += temp;
    subject_buffer2 = subject[index ];
    char* subject_buffer1_ptr = (char*) &subject_buffer1;


    n = MIN(Sequence_length - s_off, Query_length - q_off);

    char query_temp = 0, sequence_temp = 0,
                                         flag = 0;

    int temp2 = 4-(s_off&0x3);
    int n1=n-temp2;

    int n2 = n1>>2;
    int k=0;

    if(n>0)
    {
        for(k=0; k<temp2; k++)
        {
            sequence_temp = subject_buffer1_ptr[(s_off&0x3) + k];

            query_temp = query[q_off+k];

            score += matrix[ (query_temp<<5) + sequence_temp ];

            if (score > maxscore)
            {
                maxscore = score;
                best_i = k;
            }

            if (score <= 0 || (maxscore - score) >= dropoff)
            {
                flag = 1;
                break;
            }

        }

        if ( (0==flag) && (n>temp2) )
        {
            for (int i = 0; i < n2; i++)
            {

                subject_buffer1=subject_buffer2;
                index=index + temp;
                subject_buffer2 = subject[index];

                //0
                sequence_temp = subject_buffer1_ptr[0];

                query_temp = query[q_off+k];

                score += matrix[ query_temp*32 + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if (score <= 0 || (maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }

                //1
                k++;
                sequence_temp = subject_buffer1_ptr[1];

                query_temp = query[q_off+k];

                score += matrix[ query_temp*32 + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if (score <= 0 || (maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }
                //2
                k++;
                sequence_temp = subject_buffer1_ptr[2];

                query_temp = query[q_off+k];

                score += matrix[ query_temp*32 + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if (score <= 0 || (maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }

                //3
                k++;
                sequence_temp = subject_buffer1_ptr[3];

                query_temp = query[q_off+k];

                score += matrix[ query_temp*32 + sequence_temp ];

                if (score > maxscore)
                {
                    maxscore = score;
                    best_i = k;
                }

                if (score <= 0 || (maxscore - score) >= dropoff)
                {
                    flag = 1;
                    break;
                }

                k++;

            }


            if(0==flag)
            {
                subject_buffer1=subject_buffer2;
                for(int j=0; j<(n1 &0x3); j++,k++)
                {
                    sequence_temp = subject_buffer1_ptr[j];

                    query_temp = query[q_off+k];

                    score += matrix[ query_temp*32 + sequence_temp ];

                    if (score > maxscore)
                    {
                        maxscore = score;
                        best_i = k;
                    }

                    if (score <= 0 || (maxscore - score) >= dropoff)
                        break;

                }
            }

        }
    }


    ExtendRightReturn ReturnValues;
    ReturnValues.length = best_i + 1;
    ReturnValues.s_last_off = s_off + k;
    ReturnValues.maxscore = maxscore;

    return ReturnValues;
}
